#define __PROG_TYPES_COMPAT__

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>

#include "lcd.h"
#include "gfx.h"

#define UP      0
#define DOWN    1
#define LEFT    2
#define RIGHT   3
#define NONE    5

#define _SCREEN_HEIGHT 	128
#define _SCREEN_WIDTH		128

#define HEAD 1
#define BODY 2
#define TAIL 3
#define FOOD 4

#define INITIAL_SIZE    10
#define MAX_SIZE        100
#define BLOCK_SIZE      2


typedef struct {
    uint8_t x;
    uint8_t y;
} coord;


uint8_t start_game = 0;
uint8_t snake_direction;
coord snake[MAX_SIZE];
coord food;
uint32_t head_pos;
uint32_t tail_pos;
uint32_t snake_size;
char buffer[25];


void init_buttons(void)
{
    // UP
    DDRD &= ~(1 << PD6);
    PORTD |= (1 << PD6);

    // LEFT
    DDRD &= ~(1 << PD7);
    PORTD |= (1 << PD7);

    // RIGHT
    DDRD &= ~(1 << PD4);
    PORTD |= (1 << PD4);

    // DOWN
    DDRD &= ~(1 << PD5);
    PORTD |= (1 << PD5);

    // SELECT
    DDRD &= ~(1 << PD3);
    PORTD |= (1 << PD3);

}

void check_buttons(void)
{

    // UP
    if ((PIND & (1 << PD6)) == 0) {
        snake_direction = UP;
        return;
    }

    // LEFT
    if ((PIND & (1 << PD7)) == 0) {
        snake_direction = LEFT;
        return;
    }

    // RIGHT
    if ((PIND & (1 << PD4)) == 0) {
        snake_direction = RIGHT;
        return;
    }

    // DOWN
    if ((PIND & (1 << PD5)) == 0) {
        snake_direction = DOWN;
        return;
    }

    // SELECT
    if ((PIND & (1 << PD3)) == 0) {
        if(start_game == 0)
			start_game = 1;
		else
			start_game = 0;
    }
}

void init_snake(uint32_t seed)
{
    uint8_t i;

    snake_size = INITIAL_SIZE;
	snake_direction = RIGHT;
	
    for (i = 0; i < INITIAL_SIZE; ++i) {
        snake[i].x = (_SCREEN_WIDTH - INITIAL_SIZE) / 2 + i * BLOCK_SIZE;
        snake[i].y = _SCREEN_HEIGHT / 2;
    }

    head_pos = INITIAL_SIZE - 1;
    tail_pos = 0;

    srand(seed);

    // generate food
	food.x = rand() % _SCREEN_WIDTH;
	food.y = rand() % _SCREEN_HEIGHT;
}

void update_game(void)
{
    uint32_t i, prev;

    // update head
    prev = head_pos;
    head_pos = (head_pos + 1) % MAX_SIZE;

    snake[head_pos] = snake[prev];

    switch(snake_direction) {
        case UP:
			if(snake[head_pos].y <= 0) {
				start_game = 0;
				return;
			}
			else
				snake[head_pos].y = snake[head_pos].y - BLOCK_SIZE;
            break;
        case DOWN:
			if(snake[head_pos].y >= _SCREEN_HEIGHT) {
				start_game = 0;
				return;
			}
			else
				snake[head_pos].y = snake[head_pos].y + BLOCK_SIZE;
            break;
        case LEFT:
			if(snake[head_pos].x <= 0) {
				start_game = 0;
				return;
			}
			else
				snake[head_pos].x = snake[head_pos].x - BLOCK_SIZE;
            break;
        case RIGHT:
			if(snake[head_pos].x >= _SCREEN_WIDTH) {
				start_game = 0;
				return;
			}
			else
				snake[head_pos].x = snake[head_pos].x + BLOCK_SIZE;
            break;
    }
	
	// check collision
	for(i = tail_pos; i < head_pos; i++) {
		if(snake[i].x == snake[head_pos].x && snake[i].y == snake[head_pos].y) {
			start_game = 0;
			return;
		}
	}
	
    // check food
    if (abs(food.x - snake[head_pos].x) < BLOCK_SIZE && abs(food.y - snake[head_pos].y) < BLOCK_SIZE) {
        ++snake_size;
		if(snake_size > MAX_SIZE) 
			start_game = 0;
		food.x = rand() % _SCREEN_WIDTH;
		food.y = rand() % _SCREEN_HEIGHT;
    } else {
        // update tail
        tail_pos = (tail_pos + 1) % MAX_SIZE;
    }
	
}

void draw_start_screen(void)
{
    LCD_fill_screen(ST7735_GREEN);// green screen background
    GFX_set_text_color(ST7735_BLACK);
    GFX_set_cursor(30,70);
    GFX_print_str("Start Game!");
}

void draw_end_screen(void)
{
    LCD_fill_screen(ST7735_RED);// red screen background
    GFX_set_text_color(ST7735_BLACK);
    GFX_set_cursor(40,70);
    GFX_print_str("Game Over!");
	GFX_set_cursor(47, 90);
	GFX_print_str("Replay");
}

void draw_game(void)
{
    uint32_t i;
    coord block;

    // white screen background
    //LCD_fill_screen(ST7735_WHITE);
	
	// clear snake
    for (i = 0; i < snake_size; ++i) {
        block = snake[(tail_pos + i) % MAX_SIZE];
        fillRect(block.x - BLOCK_SIZE / 2, block.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE, ST7735_WHITE);
    }
	
	// clear food
	fillRect(food.x - BLOCK_SIZE / 2, food.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE, ST7735_WHITE);
	
	update_game();
	
	// draw score
	sprintf(buffer, "Score: %d", snake_size - INITIAL_SIZE);
	GFX_set_text_color(ST7735_WHITE);
    GFX_set_cursor(5,5);
    GFX_print_str(buffer);	

    // draw snake
    for (i = 0; i < snake_size; ++i) {
        block = snake[(tail_pos + i) % MAX_SIZE];
        fillRect(block.x - BLOCK_SIZE / 2, block.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE, ST7735_BLACK);
    }

    // draw food
    fillRect(food.x - BLOCK_SIZE / 2, food.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE, ST7735_BLACK);
}

int main(void)
{
    init_buttons();
    LCD_init();

    start_game = 0;
    snake_direction = NONE;
	
	init_snake(42);

while(1)
{
LCD_fill_screen(ST7735_BLACK);
}

	draw_game();

	
	goto start_screen;

    // wait for user to start the game
start_screen:
    while(1)
    {
        check_buttons();
        // start game
        if (start_game == 1) goto main_loop;

        draw_end_screen();

        _delay_ms(200);
    }
	
	// wait for user to start another game
end_screen:
    while(1)
    {
        check_buttons();

        // start game
        if (start_game == 1) {
			snake_direction = NONE;
			init_snake(42);
			//clear screen
			LCD_fill_screen(ST7735_WHITE);
			goto main_loop;
		}
		
        draw_end_screen();

        _delay_ms(200);
    }

main_loop:
    while (1)
    {
        check_buttons();
		if (start_game == 0) goto end_screen;
        draw_game();
        _delay_ms(50);
    }
	
    return 0;
}

